//==========================================================================
// File: compass.c
// Description: An implementation of compass routing
// Author: Pat Morin
// Created: Wed Jan 20 20:13:58 EST 1999
//==========================================================================
#include "compass.h"

#include <LEDA/list.h>
#include <LEDA/ray.h>

// Pi * 2 (from Knuth)
#define PI_TIMES_2 (3.1415926535897932384626433832795028841972*2.0)

// return the minimum of a and b
#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

// return the smaller angle between two directions
#define DIR_CMP(d1, d2) min(fabs((d1)-(d2)), fabs(fabs((d1)-(d2))-PI_TIMES_2))

//==========================================================================
// Return the compass successor of u when the destination is v
// Complexity: O(degree(u))
//========================================================================
inline 
edge compass_successor(graph &G, node_array<point> &pos, node u, node v) 
{
  double dir1 = ray(pos[u], pos[v]).direction();
  double dirmin = PI_TIMES_2 + 1;
  edge emin = nil;
  edge e;
  forall_adj_edges(e, u) {
    double dir2 = ray(pos[u], pos[G.target(e)]).direction();
    if (DIR_CMP(dir1, dir2) < dirmin) {
      dirmin = DIR_CMP(dir1, dir2);
      emin = e;
    }
  }
  return emin;
}

//==========================================================================
// Find the compass routing path between all pairs of vertices in G.
// Complexity: O(n^2)
// Implementation: 
// Just uses COMPASS_PATH n times.
//========================================================================
void ALL_PAIRS_COMPASS_PATHS(graph &G, node_array<point>& pos, 
			    node_matrix<double>& dist)
{
  node_array<double> adist(G);
  node_array<edge> pred(G);

  node v;
  forall_nodes (v, G) {
    COMPASS_PATH(G, v, pos, adist, pred);
    node u;
    forall_nodes (u, G)
      dist[u][v] = adist[u];
  }
}


//==========================================================================
// Find the compass routing path from all vertices of G to t
// Complexity: O(n)
// Implementation:
// Memoizes subpaths in order to achieve linear running time.  dist[u]
// is set to -1 if u has not yet been visited, is set to 0 if u has
// been visited but its distance hasn't yet been computed, and is set
// to its correct value otherwise.  This is how we detect trapping
// cycles.
//========================================================================
void COMPASS_PATH(graph &G, node t, node_array<point> &pos, 
		  node_array<double> &dist, node_array<edge> &pred)
{
  // initialize the distance array
  node u;
  forall_nodes (u, G) {
    dist[u] = -1.0;
  }
  dist[t] = 0.0;

  // initialize the only entry we need in the pred array
  pred[t] = nil;

  // find the compass path from each node u to t
  forall_nodes (u, G) {
    node cur = u; // the current node
    list<node> P; // the entire compass path from u to v
    P.push(cur);

    // keep going until we get to a node we've already visited
    while (dist[cur] < 0.0) {
      edge e = compass_successor(G, pos, cur, t);
      dist[cur] = 0.0;
      pred[cur] = e;
      P.push(G.target(e));
      cur = G.target(e);
    }

    // check if we're trapped in a cycle, and if so mark all nodes
    // on the path as being at distance infinity from v
    node prev = P.pop();
    if (prev != t && dist[prev] == 0.0) { 
      dist[prev] = HUGE_VAL;
      while (!P.empty())
	dist[P.pop()] = HUGE_VAL;
    }

    // compute the distances by working backwards in the path
    while (!P.empty()) {
      node cur = P.pop();
      dist[cur] = pos[cur].distance(pos[prev]) + dist[prev];
      prev = cur;
    }
  }
}







