//==========================================================================
// File: rcompass.c
// Description: An implementation of randomized compass routing
// Author: Pat Morin
// Created: Wed Jan 20 20:13:58 EST 1999
//==========================================================================
#include "rcompass.h"

//==========================================================================
// Return a randomized compass successor of u when the destination is v
// Complexity: O(degree(u))
//========================================================================
inline 
edge randomized_compass_successor(graph &G, node_array<point> &pos, 
				  node u, node v, random_source &S) 
{
  // find the two consecutive edges that contain v in their "cone"
  edge e = G.first_adj_edge(u);
  while (G.target(e) != v
	 && !collinear(pos[u], pos[G.target(e)], pos[v])
	 && !(left_turn(pos[u], pos[G.target(e)], pos[v]) 
	      && right_turn(pos[u], 
			    pos[G.target(G.cyclic_adj_succ(e))], 
			    pos[v])))
    e = G.cyclic_adj_succ(e);
  
  // if we are adjacent to v or one of the edges is collinear with v
  // then there is no choice - otherwise we randomly choose between
  // the two edges that contain v in their cone
  if (G.target(e) != v 
      && !collinear(pos[u], pos[G.target(e)], pos[v])) {
    bool coin;
    S >> coin;
    if (!coin)
      e = G.cyclic_adj_succ(e);
  }
  return e;
}


//==========================================================================
// Find a randomized compas routing path between all pairs of vertices
// in G.
// Complexity: unbounded, probably O(n^3) expected if G is a triangulation
// Precondition:
// G admits randomized compass routing
// Implementation:
// Implements randomized compass routing exactly as described in the
// paper.
//========================================================================
void ALL_PAIRS_RANDOMIZED_COMPASS_PATHS(graph &G, node_array<point>& pos, 
					node_matrix<double>& dist)
{
  random_source S;

  node u, v;
  forall_nodes(v, G) {
    forall_nodes(u, G) {
      dist[u][v] = 0.0;
      node cur = u;  // the current node
      // keep going until we reach v
      while (cur != v) {
	edge e = randomized_compass_successor(G, pos, cur, v, S);
	dist[u][v] += pos[G.source(e)].distance(pos[G.target(e)]);
	cur = G.target(e);
      }
    }
  }
}
